import urwid
from redial22.tree.node import Node


# TODO get rid of this if possible
class State:
    focused = None


class UITreeWidget(urwid.TreeWidget):
    indent_cols = 4

    def __init__(self, node):
        self.__super.__init__(node)
        self.key_handler = node.key_handler
        # set style for node type
        self._w = urwid.AttrWrap(self._w, node.get_value().nodetype, node.get_value().nodetype + "_focus")

    def get_display_text(self):
        if State.focused and State.focused.get_node() == self.get_node():
            return self.get_node().get_value().name + " " + self.get_node().get_value().hostinfo.ip
        else:
            return self.get_node().get_value().name

    def selectable(self):
        return True

    def keypress(self, size, key):
        if key in ("-", "left") and not self.is_leaf:
            self.expanded = False
            self.update_expanded_icon()
        elif key in ("+", "right") and not self.is_leaf:
            self.expanded = True
            self.update_expanded_icon()
        else:
            return self.key_handler(key, self)


class UITreeListBox(urwid.TreeListBox):
    def set_focus_to_node(self, node: Node):
        start = self.focus_position
        while True:
            w, n = self.body.get_next(start)
            if n is None:
                return
            if n.get_value() == node:
                self.set_focus(n)
                return
            start = n

    # todo remove duplicate code with the above method
    def collapse_node(self, node: Node):
        start = self.body.focus.get_root()
        while True:
            w, n = self.body.get_next(start)
            if n is None:
                return
            if n.get_value() == node:
                w.expanded = False
                w.update_expanded_icon()
                return
            start = n

    def keypress(self, size, key):
        if key == "left":
            widget, pos = self.body.get_focus()
            self.keypress(size, "-")
            parent_widget, parent_pos = self.body.get_focus()

            if pos == parent_pos and size[1] > 2:
                self.move_focus_to_parent(size)

        else:
            if key not in ['home', 'end']:
                self.__super.keypress(size, key)


class UITreeNode(urwid.TreeNode):
    """ Data storage object for leaf nodes """

    def __init__(self, value, parent=None, key=None, depth=None, key_handler=None):
        urwid.TreeNode.__init__(self, value, parent=parent, key=key, depth=depth)
        self.key_handler = key_handler

    def load_widget(self):
        return UITreeWidget(self)

    """
        This is done to show IP address for connection on focus.
        Doesn't seem to be efficient. 
    """
    def get_widget(self, reload=False):
        return super().get_widget(True)


class UIParentNode(urwid.ParentNode):
    """ Data storage object for interior/parent nodes """
    def __init__(self, value, parent=None, key=None, depth=None, key_handler=None):
        urwid.ParentNode.__init__(self, value, parent=parent, key=key, depth=depth)
        self.key_handler = key_handler

    def load_widget(self):
        return UITreeWidget(self)

    def load_child_keys(self):
        data = self.get_value()
        return range(len(data.children))

    def load_child_node(self, key):
        """Return either an UITreeNode or UIParentNode"""
        child_data = self.get_value().children[key]
        child_depth = self.get_depth() + 1
        if 'folder' in child_data.nodetype:
            child_class = UIParentNode
        else:
            child_class = UITreeNode
        return child_class(child_data, parent=self, key=key, depth=child_depth, key_handler=self.key_handler)




import os
from os.path import expanduser

import urwid
from redial22.hostinfo import HostInfo
from redial22.tree.node import Node
from redial22.utils import get_public_ssh_keys
from urwid.tests.util import SelectableText


def textbox(markup, label=""):
    return urwid.AttrWrap(urwid.Edit(label, markup, wrap=urwid.widget.CLIP), 'dialog_edit', 'dialog_edit_focus')


# TODO validations
# * special characters
# * identity file exists
class AddHostDialog:

    def __init__(self, parent: Node, target: Node, on_close):
        self.parent = parent
        self.target = target
        self.on_close = on_close

        self.show_advanced = False
        self.loop = None

        # Form Fields
        label_connection_name = urwid.Text("Name")

        self.connection_name = textbox(target.name)

        label_ip = urwid.Text("IP")
        self.ip = textbox(target.hostinfo.ip)

        label_username = urwid.Text("Username")
        self.username = textbox(target.hostinfo.username)

        label_port = urwid.Text("Port")
        self.port = textbox(target.hostinfo.port if self.target.name else "22")

        label_id_file = urwid.Text("Private Key Path")
        self.id_file = textbox(target.hostinfo.identity_file)

        label_dynamic_forward = urwid.Text("Dynamic Forward")
        self.dynamic_forward = textbox(target.hostinfo.dynamic_forward)

        label_local_forward = urwid.Text("Local Forward")
        self.local_forward_from = textbox(target.hostinfo.local_forward[0], "from ")
        self.local_forward_to = textbox(target.hostinfo.local_forward[1], "to ")
        self.local_forward = urwid.Columns([self.local_forward_from, self.local_forward_to])

        label_remote_forward = urwid.Text("Remote Forward")
        self.remote_forward_from = textbox(target.hostinfo.remote_forward[0], "from ")
        self.remote_forward_to = textbox(target.hostinfo.remote_forward[1], "to ")
        self.remote_forward = urwid.Columns([self.remote_forward_from, self.remote_forward_to])

        label_proxy_jump = urwid.Text("ProxyJump")
        self.proxy_jump = textbox(target.hostinfo.proxy_jump if hasattr(target.hostinfo, 'proxy_jump') else "")

        # Header
        self.header_text = 'Edit Connection' if self.target.name else 'Add Connection'

        # Advanced
        self.advanced_btn = urwid.AttrWrap(urwid.CheckBox('Advanced', False, False, self.on_advanced),
                                           'dialog_button', 'dialog_button_focus')

        # Footer
        save_btn = urwid.AttrWrap(urwid.Button('Save', self.on_save),
                                  'dialog_button', 'dialog_button_focus')

        cancel_btn = urwid.AttrWrap(urwid.Button('Cancel', self.on_cancel),
                                    'dialog_button', 'dialog_button_focus')

        self.footer = urwid.GridFlow([save_btn, cancel_btn], 12, 1, 1, 'center')

        label_pile = urwid.Pile([label_connection_name, label_ip, label_username, label_port])
        edit_pile = urwid.Pile([self.connection_name, self.ip, self.username, self.port])

        advanced_label_pile = urwid.Pile([label_connection_name, label_ip, label_username, label_port, label_id_file,
                                          label_dynamic_forward, label_local_forward, label_remote_forward, label_proxy_jump])
        advanced_edit_pile = urwid.Pile([self.connection_name, self.ip, self.username, self.port, self.id_file,
                                         self.dynamic_forward, self.local_forward, self.remote_forward, self.proxy_jump])

        self.body = urwid.Filler(
            urwid.Pile([
                urwid.Columns([('weight', 1, label_pile), ('weight', 2, edit_pile)]),
                self.advanced_btn,
                self.footer
            ])
        )

        self.advanced_body = urwid.Filler(
            urwid.Pile([
                urwid.Columns([('weight', 1, advanced_label_pile), ('weight', 2, advanced_edit_pile)]),
                self.footer
            ])
        )

    def show(self, loop):
        self.loop = loop
        w = DialogOverlay(
            on_close=lambda: self.on_close(self.parent),
            on_enter=self.on_save,
            top_w=urwid.AttrMap(
                urwid.LineBox(
                    self.advanced_body if self.show_advanced else self.body, self.header_text),
                "dialog"),
            bottom_w=loop.widget,
            align='center',
            width=64,
            valign='middle',
            height=11 if self.show_advanced else 8
        )

        loop.widget = w

    def on_advanced(self, args, user_data):
        self.show_advanced = not self.show_advanced
        self.show(self.loop)

    def on_save(self, args=None):
        host_info = HostInfo(self.connection_name.edit_text)
        host_info.ip = self.ip.edit_text
        host_info.username = self.username.edit_text
        host_info.port = self.port.edit_text
        host_info.identity_file = self.id_file.edit_text
        host_info.dynamic_forward = self.dynamic_forward.edit_text
        host_info.local_forward = (self.local_forward_from.edit_text, self.local_forward_to.edit_text)
        host_info.remote_forward = (self.remote_forward_from.edit_text, self.remote_forward_to.edit_text)
        host_info.proxy_jump = self.proxy_jump.edit_text

        self.target.name = self.connection_name.edit_text
        self.target.hostinfo = host_info

        self.parent.add_child(self.target)

        self.on_close(self.target)

    def on_cancel(self, args):
        self.on_close(self.parent)


class RemoveHostDialog:

    def __init__(self, parent: Node, target: Node, on_close):
        self.parent = parent
        self.target = target
        self.on_close = on_close

    def show(self, loop):
        # Header
        header_text = urwid.Text('Remove Connection: ' + self.target.name, align='center')
        header = urwid.AttrMap(header_text, 'dialog')

        # Footer
        ok_btn = urwid.Button('Ok', self.on_ok)
        ok_btn = urwid.AttrWrap(ok_btn, 'dialog_button', 'dialog_button_focus')

        cancel_btn = urwid.Button('Cancel', self.on_cancel)
        cancel_btn = urwid.AttrWrap(cancel_btn, 'dialog_button', 'dialog_button_focus')

        footer = urwid.GridFlow([ok_btn, cancel_btn], 12, 1, 1, 'center')

        body = urwid.Filler(
            urwid.Pile([
                urwid.Text("Are you sure?"),
                urwid.Text(""),
                footer
            ])
        )

        # Layout
        layout = urwid.Frame(
            body,
            header=header
        )

        w = DialogOverlay(
            on_close=lambda: self.on_close(self.target),
            on_enter=lambda: self.on_ok(None),
            top_w=urwid.AttrMap(urwid.LineBox(layout), "dialog"),
            bottom_w=loop.widget,
            align='center',
            width=40,
            valign='middle',
            height=10
        )

        loop.widget = w

    def on_ok(self, args):
        self.parent.remove_child(self.target)
        self.on_close(self.parent)

    def on_cancel(self, args):
        self.on_close(self.target)


class AddFolderDialog:

    def __init__(self, parent: Node, target: Node, on_close):
        self.parent = parent
        self.target = target
        self.on_close = on_close

        # Form Fields
        self.folder_name = urwid.Edit("Folder Name: ", target.name)

    def show(self, loop):
        # Header
        header_text = urwid.Text('Edit Folder' if self.target.name else "Add Folder", align='center')
        header = urwid.AttrMap(header_text, 'dialog')

        # Footer
        save_btn = urwid.Button('Save', self.on_save)
        save_btn = urwid.AttrWrap(save_btn, 'dialog_button', 'dialog_button_focus')

        cancel_btn = urwid.Button('Cancel', self.on_cancel)
        cancel_btn = urwid.AttrWrap(cancel_btn, 'dialog_button', 'dialog_button_focus')

        footer = urwid.GridFlow([save_btn, cancel_btn], 12, 1, 1, 'center')

        body = urwid.Filler(
            urwid.Pile([self.folder_name, footer])
        )

        # Layout
        layout = urwid.Frame(
            body,
            header=header)

        w = DialogOverlay(
            on_close=lambda: self.on_close(self.target),
            on_enter=self.on_save,
            top_w=urwid.AttrMap(urwid.LineBox(layout), "dialog"),
            bottom_w=loop.widget,
            align='center',
            width=40,
            valign='middle',
            height=10
        )

        loop.widget = w

    def on_save(self, args=None):
        self.target.name = self.folder_name.edit_text
        self.parent.add_child(self.target)
        self.on_close(self.target)

    def on_cancel(self, args):
        self.on_close(self.parent)


class SSHListBox(urwid.ListBox):
    def __init__(self, on_enter, **kwargs):
        self.on_enter = on_enter
        super(SSHListBox, self).__init__(**kwargs)

    def keypress(self, size, key):
        if key == "enter":
            self.on_enter()
        else:
            return super(SSHListBox, self).keypress(size, key)


class CopySSHKeyDialog:

    def __init__(self, target: Node, on_close, change_log):
        self.target = target
        self.on_close = on_close
        self.change_log = change_log

        # Form Fields
        self.ssh_keys_walker = urwid.SimpleListWalker(
            self.prepare_ssh_list_elements()
        )

    def show(self, loop):
        # Header
        header_text = urwid.Text('Select Public SSH Key to Copy', align='center')
        header = urwid.AttrMap(header_text, 'dialog')

        cancel_btn = urwid.Button('Cancel', self.on_cancel)
        cancel_btn = urwid.AttrWrap(cancel_btn, 'dialog_button', 'dialog_button_focus')

        footer = urwid.GridFlow([cancel_btn], 12, 1, 1, 'center')

        list_box = urwid.BoxAdapter(SSHListBox(on_enter=self.on_copy, body=self.ssh_keys_walker), 5)

        body = urwid.Filler(
            urwid.Pile([list_box, footer])
        )

        # Layout
        layout = urwid.Frame(
            body=body,
            header=header)

        layout = urwid.AttrWrap(layout, 'dialog')

        w = DialogOverlay(
            on_close=lambda: self.on_close(),
            on_enter=self.on_copy,
            top_w=urwid.AttrMap(urwid.LineBox(layout), "dialog"),
            bottom_w=loop.widget,
            align='center',
            width=40,
            valign='middle',
            height=10
        )

        loop.widget = w

    def on_copy(self, args=None):
        ssh_files_path = os.path.join(expanduser("~"), '.ssh/')
        selected_ssh_key = os.path.join(ssh_files_path, self.ssh_keys_walker.get_focus()[0].original_widget.text)
        command = self.target.hostinfo.get_ssh_copy_command(selected_ssh_key)
        self.on_close(command)

    def on_cancel(self, args):
        self.change_log(None)
        self.on_close()

    def prepare_ssh_list_elements(self):
        ssh_keys = []
        for key in get_public_ssh_keys():
            ssh_keys.append(urwid.AttrWrap(SelectableText(os.path.basename(key)), 'ssh_copy', 'ssh_copy_focus'))
        return ssh_keys


class MessageDialog:

    def __init__(self, title, message, on_close):
        self.title = title
        self.message = message
        self.on_close = on_close

    def show(self, loop):
        # Header
        header_text = urwid.Text(self.title, align='center')
        header = urwid.AttrMap(header_text, 'dialog')

        # Footer
        ok_btn = urwid.Button('Ok', self.on_ok)
        ok_btn = urwid.AttrWrap(ok_btn, 'dialog_button', 'dialog_button_focus')

        footer = urwid.GridFlow([ok_btn], 12, 1, 1, 'center')

        body = urwid.Filler(
            urwid.Pile([
                urwid.Text(self.message),
                urwid.Text(""),
                footer
            ])
        )

        # Layout
        layout = urwid.Frame(
            body,
            header=header
        )

        w = DialogOverlay(
            on_close=self.on_close,
            on_enter=self.on_ok,
            top_w=urwid.AttrMap(urwid.LineBox(layout), "dialog"),
            bottom_w=loop.widget,
            align='center',
            width=40,
            valign='middle',
            height=10
        )

        loop.widget = w

    def on_ok(self, args=None):
        self.on_close()


class DialogOverlay(urwid.Overlay):

    def __init__(self, on_close, on_enter, **kwargs):
        self.on_close = on_close
        self.on_enter = on_enter
        super(DialogOverlay, self).__init__(**kwargs)

    def keypress(self, size, key):
        if key == 'shift tab':
            key = 'up'
        elif key == 'tab':
            key = 'down'
        elif key == 'esc':
            self.on_close()
        # TODO: implement, save with "enter"

        super().keypress(size, key)


class RemoveFolderDialog:
    def __init__(self, parent: Node, target: Node, on_close):
        self.parent = parent
        self.target = target
        self.on_close = on_close

    def show(self, loop):
        # Header
        header_text = urwid.Text('Remove Folder: ' + self.target.name, align='center')
        header = urwid.AttrMap(header_text, 'dialog')

        # Body warning if not empty
        warning = "Folder is not empty. All contents will be removed." if self.target.children else "Are you sure?"

        # Footer
        ok_btn = urwid.Button('Ok', self.on_ok)
        ok_btn = urwid.AttrWrap(ok_btn, 'dialog_button', 'dialog_button_focus')

        cancel_btn = urwid.Button('Cancel', self.on_cancel)
        cancel_btn = urwid.AttrWrap(cancel_btn, 'dialog_button', 'dialog_button_focus')

        footer = urwid.GridFlow([ok_btn, cancel_btn], 12, 1, 1, 'center')

        body = urwid.Filler(
            urwid.Pile([
                urwid.Text(warning),
                urwid.Text(""),
                footer
            ])
        )

        # Layout
        layout = urwid.Frame(
            body,
            header=header
        )

        w = DialogOverlay(
            on_close=lambda: self.on_close(self.parent),
            on_enter=lambda: self.on_ok(None),
            top_w=urwid.AttrMap(urwid.LineBox(layout), "dialog"),
            bottom_w=loop.widget,
            align='center',
            width=40,
            valign='middle',
            height=10
        )

        loop.widget = w

    def on_ok(self, args):
        self.parent.remove_child(self.target)
        self.on_close(self.parent)

    def on_cancel(self, args):
        self.on_close(self.target)



palette = [
    ('body', 'white', 'black'),
    ('session', 'light gray', 'black'),
    ('session_focus', 'black', 'light green', 'standout'),
    ('folder', 'light blue', 'black', 'bold'),
    ('folder_focus', 'black', 'light green', 'bold'),
    ('head', 'yellow', 'black', 'standout'),
    ('foot', 'light gray', 'black'),
    ('key', 'light cyan', 'black', 'underline'),
    ('title', 'white', 'black', 'bold'),
    ('dirmark', 'black', 'dark cyan', 'bold'),
    ('flag', 'dark gray', 'light gray'),
    ('error', 'light red', 'black'),
    ('fbutton', 'white', 'dark gray'),
    ('fbutton_sc', 'light green', 'black', 'bold'),
    ('dialog', 'white', 'black'),
    ('dialog_button', 'black', 'light green'),
    ('dialog_button_focus', 'white', 'dark green', 'bold'),
    ('dialog_edit', 'black', 'light gray'),
    ('dialog_edit_focus', 'black', 'yellow'),
    ('ssh_copy', 'black', 'dark green'),
    ('ssh_copy_focus', 'white', 'dark green', 'bold'),
    #(name, foreground, background, mono, foreground_high, background_high)
]



import sys


def main():
    try:
        from .redial import run
        sys.exit(run())
    except KeyboardInterrupt:
        from . import ExitStatus
        sys.exit(ExitStatus.ERROR_CTRL_C)


if __name__ == '__main__':
    main()


import urwid


def do_nothing(key): pass


class FooterButton(urwid.Button):

    def __init__(self, sc, caption, key, callback=do_nothing):
        super(FooterButton, self).__init__("")
        self.key = key
        urwid.connect_signal(self, 'click', callback)
        self._w = urwid.AttrWrap(urwid.Text([('fbutton_sc', " " + sc + " "), " ", ('fbutton', caption)]), 'fbutton')

    def on_click(self, widget_on_focus):
        widget_on_focus[0].keypress((0, ), self.key)


def init_footer(listbox):
    connect_button = FooterButton(u"\u23ce", "Connect", "enter", lambda button: button.on_click(listbox.get_focus()))
    sshkey_button = FooterButton("F3", "Copy SSH Key", "f3", lambda button: button.on_click(listbox.get_focus()))
    mc_button = FooterButton("F5", "Browse", "f5", lambda button: button.on_click(listbox.get_focus()))
    add_folder_button = FooterButton("F6", "New Folder", "f6", lambda button: button.on_click(listbox.get_focus()))
    add_button = FooterButton("F7", "New Conn.", "f7", lambda button: button.on_click(listbox.get_focus()))
    delete_button = FooterButton("F8", "Remove", "f8", lambda button: button.on_click(listbox.get_focus()))
    edit_button = FooterButton("F9", "Edit", "f9", lambda button: button.on_click(listbox.get_focus()))
    quit_button = FooterButton("Q", "Quit", "q", lambda button: button.on_click(listbox.get_focus()))
    # TODO keys that dont depend on selected node should be handled differently

    return urwid.GridFlow([connect_button,
                           sshkey_button,
                           mc_button,
                           # TODO join add buttons to one
                           add_folder_button,
                           add_button,
                           delete_button,
                           edit_button,
                           quit_button], 18, 1, 0, 'center')



from redial22.hostinfo import HostInfo


class Node(object):
    def __init__(self, name, nodetype="folder", hostinfo: HostInfo = None):
        # TODO remove name
        self.name = name
        self.children = []
        # TODO refactor nodetype
        self.nodetype = nodetype
        self.hostinfo = hostinfo

    def add_child(self, node: 'Node'):
        child_found = [c for c in self.children if c.name == node.name]
        if child_found:
            return child_found[0]
        else:
            self.children.append(node)
            return node

    def remove_child(self, node: 'Node'):
        child_found = [c for c in self.children if c == node]
        if child_found:
            self.children.remove(child_found[0])




# -*- coding: utf-8 -*-
try:
    from importlib.metadata import version, PackageNotFoundError
    try:
        __version__ = version("redial22")
    except PackageNotFoundError:
        __version__ = 'unknown'
except ImportError:
    # Fallback for Python < 3.8
    from pkg_resources import get_distribution, DistributionNotFound
    try:
        __version__ = get_distribution("redial22").version
    except DistributionNotFound:
        __version__ = 'unknown'



import glob
import os
import shutil
from os.path import expanduser


def package_available(package_name):
    """
    Checks unix package availability
    :param package_name: Unix package name. For example: git, docker, mc
    :return: True if package is available
             False if package is not available
    """

    return True if shutil.which(package_name) is not None else False


def get_public_ssh_keys():
    """
    Gets public (pub) ssh keys from ~/.ssh directory
    :return: list of public key paths
    """
    home = expanduser("~")
    path = os.path.join(home, '.ssh/')
    files = [f for f in glob.glob(path + "**/*.pub", recursive=True)]
    return files



from redial22 import xdg
from redial22.hostinfo import HostInfo
from redial22.tree.node import Node
import os, json


class Config:

    @staticmethod
    def load_from_file():
        hosts = []
        with open(Config.__get_or_create_config_file(), "r") as file:
            host_info = None
            for f_line in file:
                line = f_line.strip()
                if line.startswith('#') or not line:
                    continue
                kv = list(filter(None, line.split(' ')))

                key = kv[0].lower()  # SSH config file keys are case insensitive

                try:
                    if key == "host":
                        if host_info is not None:
                            hosts.append(host_info)
                        if len(kv) == 1:
                            host_info = None
                        else:
                            host_info = HostInfo(kv[1])
                            host_info.proxy_jump = ""
                    if key == "hostname" and host_info:
                        host_info.ip = kv[1]
                    if key == "user" and host_info:
                        host_info.username = kv[1]
                    if key == "port" and host_info:
                        host_info.port = kv[1]
                    if key == "identityfile" and host_info:
                        host_info.identity_file = kv[1]
                    if key == "dynamicforward" and host_info:
                        host_info.dynamic_forward = kv[1]
                    if key == "localforward" and host_info:
                        host_info.local_forward = (kv[1], kv[2])
                    if key == "remoteforward" and host_info:
                        host_info.remote_forward = (kv[1], kv[2])
                    if key == "proxyjump" and host_info:
                        host_info.proxy_jump = kv[1]
                except IndexError:
                    continue

            if host_info is not None:
                hosts.append(host_info)

        return Config.__construct_tree(hosts)

    @staticmethod
    def save_to_file(sessions):
        with open(Config.__get_or_create_config_file(), "w") as file:
            Config.__append_node_to_file(sessions, "", file)

    @staticmethod
    def save_state(state: dict):
        with open(Config.__get_or_create_state_file(), "w") as file:
            json.dump(state, file, default=vars, indent=4)

    @staticmethod
    def load_state() -> dict:
        with open(Config.__get_or_create_state_file(), "r") as file:
            try:
                return json.load(file)
            except json.decoder.JSONDecodeError:
                return {}

    # Private Methods
    __CONFIG_PATH = xdg.get_config_dir()
    __CONFIG_FILE = "sessions"
    __STATE_FILE = "state.json"

    @staticmethod
    def __get_or_create_config_file() -> str:
        full_path = Config.__CONFIG_PATH + "/" + Config.__CONFIG_FILE
        if not os.path.isfile(full_path):
            file = open(full_path, "w")
            file.close()
        return full_path

    @staticmethod
    def __get_or_create_state_file() -> str:
        full_path = Config.__CONFIG_PATH + "/" + Config.__STATE_FILE
        if not os.path.isfile(full_path):
            file = open(full_path, "w")
            file.close()
        return full_path

    @staticmethod
    def __append_node_to_file(node: Node, path: str, file):
        path = path.lstrip("/")
        if node.nodetype == "folder":
            for child in node.children:
                sub_path = path + "/" + node.name if (node.name != ".") else ""
                Config.__append_node_to_file(child, sub_path, file)
        else:
            session_name = (path + "/" + node.name).lstrip("/")
            host = node.hostinfo
            file.write("Host " + session_name + "\n")
            file.write("\tHostName " + host.ip + "\n")

            if host.username:
                file.write("\tUser " + host.username + "\n")

            if host.port:
                file.write("\tPort " + host.port + "\n")

            if host.identity_file:
                file.write("\tIdentityFile " + host.identity_file + "\n")

            if host.dynamic_forward:
                file.write("\tDynamicForward " + host.dynamic_forward + "\n")

            if host.local_forward[0]:
                file.write("\tLocalForward " + host.local_forward[0] + " " + host.local_forward[1] + "\n")

            if host.remote_forward[0]:
                file.write("\tRemoteForward " + host.remote_forward[0] + " " + host.remote_forward[1] + "\n")

            if host.proxy_jump:
                file.write("\tProxyJump " + host.proxy_jump + "\n")

            file.write("\n")

    @staticmethod
    def __construct_tree(hosts):
        root = Node('.')

        for host in hosts:
            prev_part = root
            parts = host.full_name.split("/")
            for part_idx in range(len(parts)):
                if part_idx == len(parts) - 1:
                    part = prev_part.add_child(Node(parts[part_idx], "session", host))
                else:
                    part = prev_part.add_child(Node(parts[part_idx]))

                prev_part = part

        return root




from typing import Tuple

class HostInfo:
    # TODO rename to name
    full_name: str = ""
    ip: str = ""
    port: str = ""
    username: str = ""
    identity_file: str = ""
    dynamic_forward: str = ""
    local_forward: Tuple[str, str] = ("", "")
    remote_forward: Tuple[str, str] = ("", "")
    proxy_jump: str = ""

    def __init__(self, full_name: str):
        self.full_name = full_name
        self.ip = ""
        self.port = ""
        self.username = ""
        self.identity_file = ""
        self.dynamic_forward = ""
        self.local_forward = ("", "")
        self.remote_forward = ("", "")
        self.proxy_jump = ""

    def get_mc_command(self):
        c = "mc . sh://"

        if self.username:
            c = c + self.username + "@"

        c = c + self.ip

        if self.port:
            c = c + ":" + self.port

        if self.username:
            c = c + "/home/" + self.username

        return c

    def get_ssh_command(self):
        c = "ssh "

        if self.identity_file:
            c = c + "-i " + self.identity_file + " "

        if self.dynamic_forward:
            c = c + "-D" + self.dynamic_forward + " "

        if self.local_forward[0]:
            c = c + "-L" + self.local_forward[0] + ":" + self.local_forward[1] + " "

        if self.remote_forward[0]:
            c = c + "-R" + self.remote_forward[0] + ":" + self.remote_forward[1] + " "

        if self.proxy_jump:
            c = c + "-J " + self.proxy_jump + " "

        if self.username:
            c = c + self.username + "@"

        c = c + self.ip

        if self.port:
            c = c + " -p " + self.port

        return c

    def get_ssh_copy_command(self, identity_file):
        c = "ssh-copy-id -i {} ".format(identity_file)

        if self.port:
            c = c + " -p " + self.port + " "
            
        if self.username:
            c = c + self.username + "@"

        c = c + self.ip

        return c



from redial22.config import Config
from redial22.ui.tree import UITreeListBox, UITreeNode, UIParentNode

SELECTED = "selected"
COLLAPSED = "collapsed"


# todo do we really need sessions here
def restore_ui_state(listbox: UITreeListBox, sessions):
    ui_state = Config().load_state()

    if SELECTED in ui_state:
        selected = ui_state[SELECTED]
        node = find_node(sessions, selected[1:])
        listbox.set_focus_to_node(node)

    if COLLAPSED in ui_state:
        collapsed = ui_state[COLLAPSED]
        for c in collapsed:
            node = find_node(sessions, c[1:])
            listbox.collapse_node(node)


def save_ui_state(listbox: UITreeListBox):
    ui_state = dict()
    ui_state[SELECTED] = get_path(listbox.get_focus_path()[0])

    parent: UIParentNode = listbox.focus.get_node().get_root()
    ui_state[COLLAPSED] = find_collapsed(parent._children)  # todo can we avoid this

    Config().save_state(ui_state)


def find_node(node_tree: list, path: list):
    if len(path) == 0:
        return node_tree
    for node in node_tree.children:
        if node.name == path[0]:
            return find_node(node, path[1:])


def find_collapsed(nodes):
    collapsed = []
    for key in nodes:
        node = nodes[key]
        if isinstance(node, UITreeNode):
            continue
        collapsed.extend(find_collapsed(node._children))
        if node.get_widget().expanded is False:
            collapsed.append(get_path(node))
    return collapsed


def get_path(node: UITreeNode) -> list:
    path = []

    while node:
        path.append(node.get_value().name)
        node = node.get_parent()

    path.reverse()
    return path



import os
import signal
import typing

import urwid
from redial22.config import Config
from redial22.hostinfo import HostInfo
from redial22.tree.node import Node
from redial22.ui.dialog import AddHostDialog, MessageDialog, AddFolderDialog, RemoveHostDialog, CopySSHKeyDialog, RemoveFolderDialog
from redial22.ui.footer import init_footer
from redial22.ui.tree import UIParentNode, UITreeWidget, UITreeNode, UITreeListBox, State
from redial22.ui.palette import palette
from redial22.utils import package_available, get_public_ssh_keys
from redial22.uistate import save_ui_state, restore_ui_state
from functools import partial


class RedialApplication:

    def __init__(self):
        self.sessions = Config().load_from_file()

        top_node = UIParentNode(self.sessions, key_handler=self.on_key_press)
        self.walker = urwid.TreeWalker(top_node)
        self.listbox = UITreeListBox(self.walker)

        restore_ui_state(self.listbox, self.sessions)

        self.header_text = urwid.Text("redial22")
        urwid.connect_signal(self.walker, "modified", lambda: on_focus_change(self.listbox))
        footer = init_footer(self.listbox)

        self.view = urwid.Frame(
            urwid.AttrWrap(self.listbox, 'body'),
            header=urwid.AttrWrap(self.header_text, 'head'),
            footer=footer)

        # Set screen to 256 color mode
        screen = urwid.raw_display.Screen()
        screen.set_terminal_properties(256)
        self.loop = urwid.MainLoop(self.view, palette, screen)

        # instance attributes
        self.command = None
        self.command_return_key = None  # type: int | None
        self.log = None
        self.search_string = ""
        self.in_search_mode = False

    def reset_search(self):
        self.search_string = ""
        self.in_search_mode = False
        self.header_text.set_text("redial22")

    def search_and_focus(self, search_string):
        # Traverse all nodes and focus the first match
        def traverse(node, parent_path=None):
            if parent_path is None:
                parent_path = []
            results = []
            if hasattr(node, 'children') and node.children:
                for child in node.children:
                    results.extend(traverse(child, parent_path + [node]))
            results.append((node, parent_path))
            return results
        all_nodes = traverse(self.sessions)
        search_lower = search_string.lower()
        for node, parent_path in all_nodes:
            name = getattr(node, 'name', '') or ''
            nodetype = getattr(node, 'nodetype', '') or ''
            ip = getattr(getattr(node, 'hostinfo', None), 'ip', '') or ''
            if search_lower in name.lower() or search_lower in ip.lower():
                # Expand all parent folders by iterating through visible widgets
                for parent in parent_path:
                    if hasattr(parent, 'nodetype') and parent.nodetype == 'folder':
                        widget = None
                        if hasattr(parent, 'get_widget'):
                            widget = parent.get_widget()
                        if widget is not None and hasattr(widget, 'expanded'):
                            widget.expanded = True
                            widget.update_expanded_icon()
                self.listbox.set_focus_to_node(node)
                break

    def run(self):
        if self.command_return_key == 0 and self.log is not None:
            MessageDialog("Info", self.log, self.close_dialog).show(self.loop)
            self.log = None
        self.loop.run()

    def on_key_press(self, key: str, w: UITreeWidget):
        # Enter search mode with '/'
        if not self.in_search_mode and key == '/':
            self.in_search_mode = True
            self.search_string = ""
            self.header_text.set_text("Search: ")
            return

        # Handle search mode
        if self.in_search_mode:
            if key == 'esc' or key == 'enter':
                self.reset_search()
                return
            elif key == 'backspace':
                self.search_string = self.search_string[:-1]
            elif len(key) == 1 and (key.isprintable()):
                self.search_string += key
            # Update header and focus
            self.header_text.set_text(f"Search: {self.search_string}")
            self.search_and_focus(self.search_string)
            return

        # Handle reserved command keys (only when not in search mode)
        if key in ['q', 'Q', 'ctrl d']:
            self.command = EXIT_REDIAL
            raise urwid.ExitMainLoop()

        this_node = w.get_node().get_value()
        folder_node = this_node if (w.get_node().get_parent() is None or this_node.nodetype == "folder") \
            else w.get_node().get_parent().get_value()

        parent_node = None if w.get_node().get_parent() is None else w.get_node().get_parent().get_value()

        if key == "enter":
            if isinstance(w.get_node(), UITreeNode):
                self.command = w.get_node().get_value().hostinfo.get_ssh_command()
                raise urwid.ExitMainLoop()

        elif key == "f3" and w.is_leaf:
            if (len(get_public_ssh_keys())) == 0:
                MessageDialog("Error",
                              "There is no public SSH Key (.pub) in ~/.ssh folder. You can use ssh-keygen to "
                              "generate SSH key pairs",
                              self.close_dialog).show(
                    self.loop)
            else:
                self.log = "SSH key is copied successfully"
                CopySSHKeyDialog(this_node, self.close_dialog_and_run, self.change_log).show(self.loop)

        elif key == "f5" and w.is_leaf:
            if package_available(package_name="mc"):
                self.command = this_node.hostinfo.get_mc_command()
                raise urwid.ExitMainLoop()
            else:
                MessageDialog("Error", "Please install mc (Midnight Commander) package"
                                       " to use this feature", self.close_dialog).show(self.loop)

        elif key == "f6":
            AddFolderDialog(folder_node, Node("", "folder"), self.save_and_focus).show(self.loop)

        elif key == "f7":
            AddHostDialog(folder_node, Node("", "session", HostInfo("")), self.save_and_focus).show(self.loop)

        elif key == "f8":
            if parent_node is None:
                MessageDialog("Error", "Root folder cannot be removed", self.close_dialog).show(self.loop)
            else:
                parent_node_typed = typing.cast(Node, parent_node)
                if this_node.nodetype == "folder":
                    RemoveFolderDialog(parent_node_typed, this_node, self.save_and_focus).show(self.loop)
                else:
                    RemoveHostDialog(parent_node_typed, this_node, self.save_and_focus).show(self.loop)

        elif key == "f9" and w.is_leaf:
            AddHostDialog(parent_node, this_node, self.save_and_focus).show(self.loop)

        elif key in ["meta down", "ctrl down"]:
            if parent_node is None: return
            i = parent_node.children.index(this_node)
            if i == len(parent_node.children) - 1: return  # at bottom
            parent_node.children[i], parent_node.children[i + 1] = parent_node.children[i + 1], parent_node.children[i]

            self.save_and_focus(this_node)

        elif key in ["meta up", "ctrl up"]:
            if parent_node is None: return
            i = parent_node.children.index(this_node)
            if i == 0: return  # at top
            parent_node.children[i], parent_node.children[i - 1] = parent_node.children[i - 1], parent_node.children[i]

            self.save_and_focus(this_node)
        else:
            return key

    def save_and_focus(self, focus: Node):
        save_ui_state(self.listbox)
        Config().save_to_file(self.sessions)
        self.walker.set_focus(UIParentNode(self.sessions, key_handler=self.on_key_press))
        self.listbox.set_focus_to_node(focus)
        restore_ui_state(self.listbox, self.sessions)
        self.loop.widget = self.view

    def close_dialog(self):
        self.loop.widget = self.view

    def close_dialog_and_run(self, command=None):
        if command is not None:
            self.command = command
            self.loop.widget = self.view
            raise urwid.ExitMainLoop()
        else:
            self.loop.widget = self.view

    def change_log(self, log):
        self.log = log


EXIT_REDIAL = "__EXIT__"


def on_focus_change(listbox):
    State.focused = listbox.get_focus()[0]


def run():
    app = RedialApplication()

    signal.signal(signal.SIGINT, partial(sigint_handler, app))

    while True:
        app.run()

        if app.command:
            if app.command == EXIT_REDIAL:
                break
            else:
                rk = os.system(app.command)
                if rk == 33280 or rk == 0:
                    app.command_return_key = rk
                else:
                    app.command_return_key = rk
                    break

    save_ui_state(app.listbox)


def sigint_handler(app, signum, frame):
    app.command = EXIT_REDIAL
    raise urwid.ExitMainLoop()


if __name__ == "__main__":
    run()



# File copied from http-prompt source at https://github.com/eliangcs/http-prompt/blob/master/http_prompt/xdg.py

"""XDG Base Directory Specification.

See:
    https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
    https://github.com/ActiveState/appdirs
"""
import os
import sys

from functools import partial


def _get_dir(envvar_name, default_dir, resource_name=None):
    base_dir = os.getenv(envvar_name) or default_dir
    app_dir = os.path.join(base_dir, 'redial22')
    if not os.path.exists(app_dir):
        os.makedirs(app_dir, mode=0o700)

    if resource_name:
        app_dir = os.path.join(app_dir, resource_name)
        if not os.path.exists(app_dir):
            os.mkdir(app_dir)

    return app_dir


if sys.platform == 'win32':  # nocover
    # NOTE: LOCALAPPDATA is not available on Windows XP
    get_data_dir = partial(_get_dir, 'LOCALAPPDATA',
                           os.path.expanduser('~/AppData/Local'))
    get_config_dir = partial(_get_dir, 'LOCALAPPDATA',
                             os.path.expanduser('~/AppData/Local'))
else:
    get_data_dir = partial(_get_dir, 'XDG_DATA_HOME',
                           os.path.expanduser('~/.local/share'))
    get_config_dir = partial(_get_dir, 'XDG_CONFIG_HOME',
                             os.path.expanduser('~/.config'))